from typing import List, Optional

Table = List[Optional[List[str]]]


def pad(s: str, width: int) -> str:
    return " " * (width - len(s)) + s


def render_table(table: Table) -> List[str]:
    COL_DELIMITER = ' | '
    ROW_DELIMITER = '-'

    value_cell_lengths = [len(row) for row in table if row]
    if not value_cell_lengths:
        # Table is empty or contains delimiters only
        return ROW_DELIMITER * len(table)
    num_cols = max(value_cell_lengths)

    # Calc column widths
    col_widths = [0] * num_cols
    for row in table:
        if row:
            for idx, cell in enumerate(row):
                col_widths[idx] = max(col_widths[idx], len(cell))

    # Calc overall width
    width = sum(col_widths) + len(COL_DELIMITER) * (len(col_widths) - 1)

    # Output
    lines = []
    for row in table:
        line = []
        if row:
            for idx, cell in enumerate(row):
                line.append(pad(cell, col_widths[idx]))
            lines.append(COL_DELIMITER.join(line))
        else:
            lines.append(ROW_DELIMITER * width)

    return lines


def print_table(table: Table):
    lines = render_table(table)
    print("\n".join(lines))


class ConfusionMatrix:

    def __init__(self) -> None:
        self.valid_accepted = 0
        self.invalid_accepted = 0
        self.valid_rejected = 0
        self.invalid_rejected = 0

    def to_table(self) -> Table:
        total = self.valid_accepted + self.valid_rejected + self.invalid_accepted + self.invalid_rejected
        return [
            ['', 'Valid', 'Invalid', 'Total'],
            None,
            ['Accepted', str(self.valid_accepted), str(self.invalid_accepted), str(self.valid_accepted + self.invalid_accepted)],
            ['Rejected', str(self.valid_rejected), str(self.invalid_rejected), str(self.valid_rejected + self.invalid_rejected)],
            None,
            ['Total', str(self.valid_accepted + self.valid_rejected), str(self.invalid_accepted+self.invalid_rejected), str(total)],
        ]

    def render(self):
        render_table(self.to_table())

    def print(self):
        print_table(self.to_table())



from fences.core.exception import ParseException
from .xpath import NormalizedXPath

class XmlSchemaException(ParseException):

    def __init__(self, message: str, path: NormalizedXPath) -> None:
        super().__init__(f"{message} at {path}")
        self.path = path



class FencesException(Exception):
    """
    Base class of all exceptions thrown by fences.
    """
    pass


class ResolveReferenceException(FencesException):
    """
    Thrown by Node.resolve()
    """

    def __init__(self, message: str, node_id: str) -> None:
        super().__init__(message)
        self.node_id = node_id


class ParseException(FencesException):
    """
    Base class for all exceptions occurring during schema parsing
    """
    pass


class InternalException(FencesException):
    """
    Raised if fences reaches an invalid internal state.
    This hopefully never happens.
    """
    pass


class ConfigException(FencesException):
    """
    All config related exceptions
    """
    pass


class NormalizationException(FencesException):
    pass


class JsonPointerException(FencesException):
    pass



from typing import List, Union, Set, Dict, Tuple
from fences.core.exception import NormalizationException
from fences.json_schema.json_pointer import JsonPointer
import math
import hashlib
import json
import copy


SchemaType = Union[dict, bool]

NORM_FALSE = {'anyOf': [{'enum': []}]}
NORM_TRUE = {'anyOf': [{}]}

# This list does not contain 'integer' on purpose, see _simplify_type
ALL_TYPES = [
    'number',
    'boolean',
    'string',
    'null',
    'object',
    'array'
]


class Resolver:

    def __init__(self, schema: SchemaType):
        self.schema = schema
        self.cache = {}

    def resolve(self, pointer: JsonPointer) -> SchemaType:
        return pointer.lookup(self.schema)

    def add_normalized_schema(self, pointer: JsonPointer):
        key = str(pointer)
        assert key not in self.cache


def _invert_type(t: Union[str, List[str]]):
    return {'type': list(set(ALL_TYPES) - set(t))}


def _invert_properties(props: dict):
    new_props = {name: {'not': prop} for name, prop in props.items()}
    return {
        'type': 'object',
        'properties': new_props,
        'required': list(props.keys())
    }

def _invert_items(items: dict):
    return {
        'type': 'array',
        'items': {'not': items}
    }

_inverters = {
    'minimum': lambda x: {'type': ['number'], 'exclusiveMaximum': x},
    'maximum': lambda x: {'type': ['number'], 'exclusiveMinimum': x},
    'exclusiveMinimum': lambda x: {'type': ['number'], 'maximum': x},
    'exclusiveMaximum': lambda x: {'type': ['number'], 'minimum': x},
    'type': _invert_type,
    'enum': lambda x: {'NOT_enum': x},
    'NOT_enum': lambda x: {'enum': x},
    'maxLength': lambda x: {'type': ['string'], 'minLength': x},
    'minLength': lambda x: {'type': ['string'], 'maxLength': x},
    'properties': _invert_properties,
    'multipleOf': lambda x: {'type': ['number'], 'NOT_multipleOf': x},
    'required': lambda x: {'type': ['object'], 'properties': {i: False for i in x}},
    #'required': lambda x: {'type': ['object']},
    'items': _invert_items,
    'minItems': lambda x: {'type': 'array', 'maxItems': x},
    'pattern': lambda x: {'type': 'string', 'pattern': f"!({x})"}
}


def _invert(trivial_schema: dict) -> dict:
    # NOT(a and b and c)
    # = NOT(a) or NOT(b) or NOT(c)
    result = []
    if len(trivial_schema) == 0:
        return NORM_FALSE.copy()
    for key, value in trivial_schema.items():
        if key in _inverters:
            result.append(_inverters[key](value))
        else:
            raise NormalizationException(f"Cannot invert {key}")
    return {'anyOf': result}


def invert(norm_schema: dict, full_merge: bool) -> dict:
    return merge([
        _invert(i)
        for i in norm_schema['anyOf']
    ], full_merge)


def _merge_type(a: Union[List[str], str], b: Union[List[str], str]) -> List[str]:
    if isinstance(a, str):
        a = [a]
    if isinstance(b, str):
        b = [b]
    return list(set(a) & set(b))

def _merge_enums(a: List[any], b: List[any]) -> List[any]:
    # TODO: we assume string lists, we is not generic
    a = set(a)
    b = set(b)
    return list(a & b)

def _float_gcd(a, b, rtol = 1e-05, atol = 1e-08):
    t = min(abs(a), abs(b))
    while abs(b) > rtol * t + atol:
        a, b = b, a % b
    return a

_simple_mergers = {
    'required': lambda a, b: list(set(a) | set(b)),
    'multipleOf': lambda a, b: abs(a*b) // _float_gcd(a, b),
    'items': lambda a, b: {'allOf': [a, b]},
    'minimum': lambda a, b: max(a, b),
    'maximum': lambda a, b: min(a, b),
    'type': _merge_type,
    'minItems': lambda a, b: max(a, b),
    'maxItems': lambda a, b: min(a, b),
    'pattern': lambda a, b: f"({a})&({b})",
    'minLength': lambda a, b: max(a, b),
    'maxLength': lambda a, b: min(a, b),
    'enum': lambda a, b: a + b,
    'format': lambda a, b: a,  # todo
    'deprecated': lambda a, b: a or b,
    'NOT_enum': lambda a, b: a + b,
    'enum': _merge_enums,
}


def _merge_properties(result: dict, to_add: dict) -> dict:
    # Must merge properties and additionalProperties together
    # Schema 1: a: 1a,    b: 1b,              ...: 1n
    # Schema 2:           b: 2b,    c: 2c,    ...: 2n
    # Result:   a: 1a+2n, b: 1b+2b, c: 2c+1n, ...: 1n+2n

    props_result = result.get('properties', {})
    props_result = copy.deepcopy(props_result) # TODO: why?
    props_to_add = to_add.get('properties', {})
    additional_result = result.get('additionalProperties')
    additional_to_add = to_add.get('additionalProperties')
    for property_name, schema in props_result.items():
        if property_name in props_to_add:
            props_result[property_name] = {'allOf': [
                schema, props_to_add[property_name]
            ]}
        else:
            if additional_to_add is None:
                props_result[property_name] = schema
            else:
                props_result[property_name] = {'allOf': [
                    schema, additional_to_add
                ]}
    for property_name, schema in props_to_add.items():
        if property_name not in props_result:
            if additional_result is None:
                props_result[property_name] = schema
            else:
                props_result[property_name] = {'allOf': [
                    schema, additional_result
                ]}
    return props_result


def _merge_prefix_items(result: dict, to_add: dict) -> dict:
    # Must merge items and prefixItems together
    # Schema 1: a,   a,   a,   b...
    # Schema 2: c,   c,   d...
    # Result:   a+c, a+c, a+d, b+d...

    items_a = result.get('items', NORM_TRUE.copy())
    prefix_items_a = result.get('prefixItems', [])
    items_b = to_add.get('items', NORM_TRUE.copy())
    prefix_items_b = to_add.get('prefixItems', [])
    assert isinstance(prefix_items_a, list)
    assert isinstance(prefix_items_b, list)
    result_prefix_items = []

    if len(prefix_items_a) > len(prefix_items_b):
        prefix_items_b += [items_b] * \
            (len(prefix_items_a) - len(prefix_items_b))
    else:
        prefix_items_a += [items_a] * \
            (len(prefix_items_b) - len(prefix_items_a))

    assert len(prefix_items_a) == len(prefix_items_b)
    for i, j in zip(prefix_items_a, prefix_items_b):
        result_prefix_items.append({'allOf': [i, j]})

    return result_prefix_items


_complex_mergers = {
    'prefixItems': _merge_prefix_items,
    'properties': _merge_properties,
}


def _merge(result: dict, to_add: dict) -> None:

    for key, merger in _complex_mergers.items():
        if key in result or key in to_add:
            result[key] = merger(result, to_add)

    for key, value in result.items():
        if key not in to_add:
            # nothing to merge
            continue
        if key in _simple_mergers:
            merger = _simple_mergers[key]
            result[key] = merger(value, to_add[key])
        elif key in _complex_mergers:
            continue
        else:
            raise NormalizationException(f"Do not know how to merge '{key}'")

    # Copy all remaining keys
    for key, value in to_add.items():
        if key not in result and key not in _complex_mergers:
            result[key] = value


def merge(schemas: List[SchemaType], full_merge: bool) -> SchemaType:
    if full_merge:
        return merge_full(schemas)
    else:
        return merge_simple(schemas)


def merge_simple(schemas: List[SchemaType]) -> SchemaType:
    assert len(schemas) > 0
    results = []
    num = max([len(s['anyOf']) for s in schemas])
    for idx in range(num):
        result = {}
        for schema in schemas:
            ao = schema['anyOf']
            if len(ao) == 0:
                continue
            option = ao[idx % len(ao)]
            _merge(result, option)
        results.append(result)
    return {'anyOf': results}


def merge_full(schemas: List[SchemaType]) -> SchemaType:
    assert len(schemas) > 0
    result = [{}]
    for schema in schemas:
        new_result = []
        for option in schema['anyOf']:
            for i in result:
                ii = i.copy()
                _merge(ii, option)
                new_result.append(ii)
        result = new_result
    return {'anyOf': result}


def _simplify_type(schema: dict):
    if 'type' not in schema:
        return schema
    side_schema = schema.copy()
    types = schema.pop('type')
    if not isinstance(types, list):
        types = [types]
    types: Set[str] = set(types)

    # number contains integer
    if 'number' in types:
        if 'integer' in types:
            types.remove('integer')
    else:
        if 'integer' in types:
            types.remove('integer')
            types.add('number')
            side_schema['type'] = list(types)
            return {'allOf': [
                {'multipleOf': 1},
                side_schema
            ]}
    side_schema['type'] = list(types)
    return side_schema


def _simplify_if_then_else(schema: dict):
    # This is valid iff:
    # (not(IF) or THEN) and (IF or ELSE)
    # <==>
    #   ( IF and THEN ) or
    #   ( not(IF) and ELSE)

    # See https://json-schema.org/understanding-json-schema/reference/conditionals.html#implication
    if 'if' not in schema and 'then' not in schema and 'else' not in schema:
        return schema
    side_schema = schema.copy()
    if_schema = side_schema.pop('if', None)
    then_schema = side_schema.pop('then', None)
    else_schema = side_schema.pop('else', None)

    if if_schema is None:
        return side_schema

    if then_schema is None and else_schema is None:
        return {}

    any_of = []

    if then_schema is None:
        then_schema = {}
    any_of.append({'allOf': [
        if_schema,
        then_schema
    ]})

    if else_schema is None:
        else_schema = {}
    any_of.append({'allOf': [
        {'not': if_schema},
        else_schema
    ]})

    result = {'allOf': [
        side_schema,
        {'anyOf': any_of}
    ]}
    return result

def _simplify_const(schema: dict) -> dict:
    if 'const' not in schema:
        return schema
    schema = schema.copy()
    const = schema.pop('const')
    if 'enum' in schema:
        schema['enum'] += [const]
    else:
        schema['enum'] = [const]
    return schema

def _simplify_dependent_required(schema: dict) -> dict:
    if 'dependentRequired' not in schema:
        return schema
    # dependentRequired:
    #   a: ["b"]
    #
    # a exists | b exists | valid
    # no       | no       | yes
    # yes      | no       | no
    # no       | yes      | yes
    # yes      | yes      | yes
    # 
    # anyOf:
    # - { "properties": {"a": False, "b": True }, "required": [] }
    # - { "properties": {"a": True,  "b": True }, "required": ["a", "b"]}

    schema = schema.copy()
    dependent_required = schema.pop('dependentRequired')
    options = []
    for property, requires in dependent_required.items():
        props1 = {name: True for name in requires}
        props1[property] = False

        props2 = {name: True for name in requires}
        props2[property] = True
        options.append({"anyOf": [
            {"properties": props1},
            {"properties": props2, "required": requires + [property]}
        ]})
    return {"allOf": [schema] + options }

def _inline_refs(schema: dict, resolver: Resolver) -> Tuple[dict, bool]:
    if schema is False:
        return NORM_FALSE.copy(), False

    if schema is True:
        return NORM_TRUE.copy(), False

    contains_refs = False

    if '$ref' in schema:
        side_schema = schema.copy()
        del side_schema['$ref']
        pointer = JsonPointer.from_string(schema['$ref'])
        ref_schema = resolver.resolve(pointer)
        # We need a deepcopy here, otherwise we cannot stringify it due to circular references
        ref_schema = copy.deepcopy(ref_schema)
        schema = {'allOf': [side_schema, ref_schema]}
        contains_refs = True

    for kw in ['anyOf', 'allOf', 'oneOf']:
        for idx, sub_schema in enumerate(schema.get(kw, [])):
            (new_schema, new_contains_refs) = _inline_refs(sub_schema, resolver)
            schema[kw][idx] = new_schema
            contains_refs = contains_refs or new_contains_refs
    for kw in ['not', 'if', 'then', 'else']:
        if kw in schema:
            (new_schema, new_contains_refs) = _inline_refs(
                schema[kw], resolver)
            schema[kw] = new_schema
            contains_refs = contains_refs or new_contains_refs

    return schema, contains_refs


def _to_dnf(schema: dict, full_merge: bool) -> dict:

    if schema is False:
        return NORM_FALSE.copy()

    if schema is True:
        return NORM_TRUE.copy()

    schema = copy.deepcopy(schema)
    schema = _simplify_const(schema)
    schema = _simplify_if_then_else(schema)
    schema = _simplify_type(schema)
    schema = _simplify_dependent_required(schema)

    # anyOf
    if 'anyOf' in schema:
        any_ofs = []
        for sub_schema in schema['anyOf']:
            normalized_sub_schema = _to_dnf(
                sub_schema, full_merge)
            any_ofs.extend(normalized_sub_schema['anyOf'])
    else:
        any_ofs = [{}]

    # oneOf
    if 'oneOf' in schema:
        one_ofs = []
        normalized_sub_schemas = [
            _to_dnf(sub_schema, full_merge)
            for sub_schema in schema['oneOf']
        ]
        for idx, _ in enumerate(normalized_sub_schemas):
            options = merge([
                invert(i, full_merge) if sub_idx == idx else i
                for sub_idx, i in enumerate(normalized_sub_schemas)
            ], full_merge)
            one_ofs.extend(options['anyOf'])
    else:
        one_ofs = [{}]

    # allOf
    all_ofs = []
    side_schema = schema.copy()
    for sub_schema in ['allOf', 'anyOf', 'oneOf', 'not']:
        if sub_schema in side_schema:
            del side_schema[sub_schema]
    all_ofs.append({'anyOf': [side_schema]})
    for sub_schema in schema.get('allOf', []):
        all_ofs.append(_to_dnf(sub_schema, full_merge))

    # not
    if 'not' in schema:
        norm_schema = _to_dnf(schema['not'], full_merge)
        all_ofs.append(invert(norm_schema, full_merge))

    s = merge(all_ofs, full_merge)

    result = merge([
        {'anyOf': any_ofs},
        {'anyOf': one_ofs},
        s
    ], full_merge)
    return result


def _normalize(schema: dict, resolver: Resolver, new_refs: Dict[str, dict], full_merge: bool) -> dict:
    if schema is False:
        return NORM_FALSE.copy()

    if schema is True:
        return NORM_TRUE.copy()

    if len(schema) == 0:
        return NORM_TRUE.copy()

    # Check cache (to avoid stack overflows due to recursive schemas)
    new_ref_name = hashlib.sha1(json.dumps(schema).encode()).hexdigest()
    if new_ref_name in new_refs:
        return {'anyOf': [{'$ref': f"#/$defs/{new_ref_name}"}]}

    # Inline all references (if any)
    (schema, contains_refs) = _inline_refs(schema, resolver)

    result = _to_dnf(schema, full_merge)

    # Store new schema if sub-schemas later try to reference it
    if contains_refs:
        new_refs[new_ref_name] = result

    # Iterate sub-schemas
    for sub_schema in result['anyOf']:
        for kw in ['additionalProperties', 'items', 'additionalItems', 'contains']:
            if kw in sub_schema:
                sub_schema[kw] = _normalize(
                    sub_schema[kw], resolver, new_refs, full_merge)

        props: dict = sub_schema.get('properties', {})
        for name, sub_sub_schema in props.items():
            props[name] = _normalize(
                sub_sub_schema, resolver, new_refs, full_merge)

        prefix_items: list = sub_schema.get('prefixItems', [])
        for idx, sub_sub_schema in enumerate(prefix_items):
            prefix_items[idx] = _normalize(
                sub_sub_schema, resolver, new_refs, full_merge)

    # Return
    if contains_refs:
        return {'anyOf': [{'$ref': f"#/$defs/{new_ref_name}"}]}
    else:
        return result


def normalize(schema: SchemaType, full_merge: bool = True) -> any:
    if schema is False:
        return NORM_FALSE.copy()

    if schema is True:
        return NORM_TRUE.copy()

    if not isinstance(schema, dict):
        raise NormalizationException(
            f"Schema must be of type bool or dict, got {type(schema)}")

    # TODO: deepcopy really needed?
    new_schema = copy.deepcopy(schema)
    for kw in ['$schema', '$defs']:
        if kw in schema:
            del new_schema[kw]
    resolver = Resolver(schema)
    new_refs: Dict[str, dict] = {}
    new_schema = _normalize(new_schema, resolver, new_refs, full_merge)
    if '$schema' in schema:
        new_schema['$schema'] = schema['$schema']
    new_schema['$defs'] = new_refs
    return new_schema


def check_normalized(schema: SchemaType) -> None:
    resolver = Resolver(schema)
    checked_refs = set()
    _check_normalized(schema, resolver, checked_refs)


def _check_normalized(schema: SchemaType, resolver: Resolver, checked_refs: Set[str]) -> None:
    if not isinstance(schema, dict):
        raise NormalizationException(f"Must be a dict, got {schema}")

    keys = set(schema.keys())
    for i in ['$schema', '$defs']:
        if i in keys:
            keys.remove(i)
    if len(keys) != 1 or 'anyOf' not in keys:
        raise NormalizationException(
            f"Schema has one key not being anyOf, got {keys}")

    any_of = schema['anyOf']
    if not isinstance(any_of, list):
        raise NormalizationException(f"anyOf must be a list, is '{any_of}'")

    for idx, sub_schema in enumerate(any_of):

        # Disallowed keywords in any-of elements
        for i in ['anyOf', 'allOf', 'oneOf', 'not', 'if', 'then', 'else', 'const']:
            if i in sub_schema:
                raise NormalizationException(
                    f"'{i}' not allowed in normalized sub_schema {idx}")

        if '$ref' in sub_schema:
            if len(sub_schema.keys()) != 1:
                raise NormalizationException(
                    f"Sub-schema {idx} has other keys beside $ref")
            ref = sub_schema['$ref']
            if ref not in checked_refs:
                checked_refs.add(ref)
                pointer = JsonPointer.from_string(ref)
                _check_normalized(resolver.resolve(
                    pointer), resolver, checked_refs)

        # Traverse sub-schemas
        for kw in ['additionalProperties', 'items', 'additionalItems', 'contains']:
            if kw in schema:
                _check_normalized(schema[kw], resolver, checked_refs)

        for i in sub_schema.get('properties', {}).values():
            _check_normalized(i, resolver, checked_refs)

        for i in sub_schema.get('prefixItems', []):
            _check_normalized(i, resolver, checked_refs)



from dataclasses import dataclass
import random

from fences.regex.parse import parse
from fences.core.exception import InternalException

from typing import Optional


@dataclass
class StringProperties:
    min_length: int = 0
    max_length: int = None
    pattern: Optional[str] = None


def generate_random_string(properties: StringProperties) -> str:
    if properties.max_length is not None:
        assert properties.min_length <= properties.max_length
    if properties.pattern is None:
        return "x" * properties.min_length
    else:
        graph = parse(str(properties.pattern))
        result = None
        for i in graph.generate_paths():
            if i.is_valid:
                result = graph.execute(i.path)
                break
        if result is None:  # pragma: no cover
            raise InternalException(
                f"Failed to generate string for '{properties.pattern}'")
        if properties.max_length is not None and len(result) > properties.max_length:
            raise InternalException(
                f"Cannot generate a string for {properties.pattern} with is not longer than {properties.max_length}")
        padding = max(properties.min_length - len(result), 0)
        return "x" * padding + result


def generate_random_number(min_value: Optional[int] = None, max_value: Optional[int] = None):
    if min_value is None:
        min_value = -1000
    if max_value is None:
        max_value = +1000
    assert min_value <= max_value
    return random.randint(min_value, max_value)


def generate_random_format(format: str) -> str:
    # From https://json-schema.org/understanding-json-schema/reference/string#built-in-formats
    samples = {
        "date-time": "2018-11-13T20:20:39+00:00",
        "time": "20:20:39+00:00",
        "date": "2018-11-13",
        "duration": "P3D",
        "email": "test@example.com",
        "hostname": "example.com",
        "ipv4": "127.0.0.1",
        "ipv6": "2001:db8::8a2e:370:7334",
        "uuid": "3e4666bf-d5e5-4aa7-b8ce-cefe41c7568a",
    }
    return samples.get(format, "")



from typing import List
from fences.core.node import Node, Leaf, Node, Decision, Reference, NoOpDecision, NoOpLeaf
from .types import Grammar, Terminal, NonTerminal, RightHandSide, Concatenation, CharacterRange, Alternative, Repetition
from .exceptions import GrammarException
import string


class CreateInput(Decision):
    def description(self):
        return "Create Input"

    def apply(self, data: any) -> any:
        return []


class FetchOutput(Leaf):
    def __init__(self) -> None:
        super().__init__(None, True)

    def description(self):
        return "Fetch Output"

    def apply(self, data: list) -> any:
        return "".join(data)


class AppendString(Leaf):
    def __init__(self, is_valid: bool, string: str) -> None:
        super().__init__(None, is_valid)
        self.string = string

    def apply(self, data: list) -> any:
        data.append(self.string)
        return data

    def mask(self, s: str) -> str:
        allowed = string.ascii_letters + string.digits
        return "".join([
            (i if i in allowed else f"chr({ord(i)})") for i in s
        ])

    def description(self) -> str:
        return f"Append {self.mask(self.string)}"


def convert(grammar: Grammar, start='start') -> Node:
    if isinstance(start, NonTerminal):
        start = start.name

    all_nodes: List[Node] = []
    for non_terminal, rhs in grammar.items():
        node = NoOpDecision(non_terminal.name, False)
        node.add_transition(_convert(rhs))
        all_nodes.append(node)

    root = CreateInput(None, True)
    root.add_transition(Reference(None, start))
    root.add_transition(FetchOutput())
    root.resolve(all_nodes)
    root.optimize()

    return root


def _convert_terminal(terminal: Terminal) -> Node:
    return AppendString(True, terminal.value)


def _convert_non_terminal(non_terminal: NonTerminal) -> Node:
    return Reference(None, non_terminal.name)


def _convert_concatenation(concatenation: Concatenation) -> Node:
    sub_root = NoOpDecision(None, True)
    for i in concatenation.elements:
        sub_root.add_transition(_convert(i))
    return sub_root


def _convert_character_range(char_range: CharacterRange) -> Node:
    sub_root = NoOpDecision(None, False)

    # lower bound
    sub_root.add_transition(AppendString(True, chr(char_range.start)))

    # upper bound
    sub_root.add_transition(AppendString(True, chr(char_range.stop)))

    return sub_root


def _convert_alternative(elements: List[RightHandSide]) -> Node:
    root = NoOpDecision(None, False)
    for element in elements:
        root.add_transition(_convert(element))
    return root


def _convert_repetition(repetition: Repetition) -> Node:
    root = NoOpDecision(None, False)
    child = _convert(repetition.element)

    if repetition.start == 0:
        root.add_transition(NoOpLeaf(None, True))
    else:
        lower = NoOpDecision(None, True)
        for _ in range(repetition.start):
            lower.add_transition(child)
        root.add_transition(lower)

    if repetition.start != repetition.stop:
        if repetition.stop is None:
            stop = repetition.start + 3
        else:
            stop = repetition.stop
        upper = NoOpDecision(None, True)
        for _ in range(stop):
            upper.add_transition(child)
        root.add_transition(upper)

    return root


def _convert(rhs: RightHandSide) -> Node:
    if isinstance(rhs, Terminal):
        return _convert_terminal(rhs)
    elif isinstance(rhs, str):  # shorthand version of above
        return _convert_terminal(Terminal(rhs))
    elif isinstance(rhs, NonTerminal):
        return _convert_non_terminal(rhs)
    elif isinstance(rhs, Concatenation):
        return _convert_concatenation(rhs)
    elif isinstance(rhs, CharacterRange):
        return _convert_character_range(rhs)
    elif isinstance(rhs, Alternative):
        return _convert_alternative(rhs.elements)
    elif isinstance(rhs, list):  # shorthand version of above
        return _convert_alternative(rhs)
    elif isinstance(rhs, Repetition):
        return _convert_repetition(rhs)
    else:
        raise GrammarException(f"Unknown type in grammar: {rhs}")



from typing import Set, Dict, List, Optional, Union

from .exceptions import JsonSchemaException
from .config import Config
from .json_pointer import JsonPointer
from ..core.random import generate_random_string, StringProperties, generate_random_format
from .normalize import normalize

from fences.core.node import Decision, Leaf, Node, Reference, NoOpLeaf, NoOpDecision

from dataclasses import dataclass


@dataclass
class KeyReference:
    ref: Union[dict, list]
    key: Optional[str] = None
    has_value: bool = False

    def set(self, value):
        self.ref[self.key] = value
        self.has_value = True


class SetValueLeaf(Leaf):

    def __init__(self, id: str, is_valid: bool, value: any) -> None:
        super().__init__(id, is_valid)
        self.value = value

    def apply(self, data: KeyReference) -> any:
        data.set(self.value)
        return self.value

    def description(self) -> str:
        return f"Set to value '{self.value}'"


class InsertKeyNode(Decision):
    def __init__(self, id: str, key: str) -> None:
        super().__init__(id)
        self.key = key

    def apply(self, data: any) -> KeyReference:
        return KeyReference(data, self.key)

    def description(self) -> str:
        return f"Insert key '{self.key}'"


class CreateArrayNode(Decision):
    def apply(self, data: KeyReference) -> any:
        # Needed to handle allOf properly
        if data.has_value:
            return data.ref[data.key]
        value = []
        data.set(value)
        return value

    def description(self) -> str:
        return "Create empty array"


class AppendArrayItemNode(Decision):
    def apply(self, data: any) -> KeyReference:
        data.append(None)
        return KeyReference(data, len(data)-1)

    def description(self) -> str:
        return "Append item"


class CreateObjectNode(Decision):
    def apply(self, data: KeyReference) -> KeyReference:
        value = {}
        data.set(value)
        return value

    def description(self) -> str:
        return "Create empty dict"


class CreateInputNode(Decision):
    def __init__(self) -> None:
        super().__init__(None, False)

    def apply(self, data: any) -> any:
        return KeyReference({}, '')

    def description(self) -> Optional[str]:
        return "Create input"


class FetchOutputNode(Leaf):
    def __init__(self) -> None:
        super().__init__(None, True)

    def apply(self, data: KeyReference) -> any:
        return data.ref[data.key]

    def description(self) -> Optional[str]:
        return "Fetch output"


def default_config():
    return Config(
        key_handlers={
            'enum': parse_enum,
            'NOT_enum': parse_not_enum,
            '$ref': parse_reference,
        },
        type_handlers={
            'object': parse_object,
            'string': parse_string,
            'array': parse_array,
            'boolean': parse_boolean,
            'number': parse_number,
            'null': parse_null,
        },
        default_samples={
            'string': ["string"],
            'number': [42],
            'null': [None],
            'boolean': [True, False],
            'object': [{}],
            'array': [[]]
        },
        normalize=True,
    )


_NO_DEFAULT = object()


def _read_typesafe(data: dict, key: str, unparsed_keys: Set[str], type, typename, pointer: JsonPointer, default: any):
    if key not in data:
        if default is _NO_DEFAULT:
            raise JsonSchemaException(f"'{key}' is missing at {pointer}", None)
        else:
            return default
    value = data[key]
    if not isinstance(value, type):
        raise JsonSchemaException(f"{value} is not a {typename} at {pointer}", None)
    unparsed_keys.remove(key)
    return value


def _read_float(data: dict, key: str, unparsed_keys: Set[str], pointer: JsonPointer, default=_NO_DEFAULT) -> str:
    return _read_typesafe(data, key, unparsed_keys, (int, float), 'float', pointer, default)


def _read_int(data: dict, key: str, unparsed_keys: Set[str], pointer: JsonPointer, default=_NO_DEFAULT) -> str:
    v = _read_typesafe(data, key, unparsed_keys, (int, float), 'int', pointer, default)
    if v is default:
        return v
    if isinstance(v, float):
        if int(v) != v:
            raise JsonSchemaException(f"{v} is not a 'int' at {pointer}", None)
        v = int(v)
    return v


def _read_string(data: dict, key: str, unparsed_keys: Set[str], pointer: JsonPointer, default=_NO_DEFAULT) -> str:
    return _read_typesafe(data, key, unparsed_keys, str, 'string', pointer, default)


def _read_dict(data: dict, key: str, unparsed_keys: Set[str], pointer: JsonPointer, default=_NO_DEFAULT) -> dict:
    return _read_typesafe(data, key, unparsed_keys, dict, 'dict', pointer, default)


def _read_list(data: dict, key: str, unparsed_keys: Set[str], pointer: JsonPointer, default=_NO_DEFAULT) -> list:
    return _read_typesafe(data, key, unparsed_keys, list, 'list', pointer, default)


def parse_const(data, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    value = data['const']
    unparsed_keys.remove('const')

    root = NoOpDecision(str(path), False)
    root.add_transition(SetValueLeaf(None, True, value))
    root.add_transition(SetValueLeaf(None, False, f"{value}_INvALID"))
    return root


def parse_not_enum(data: dict, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    values = _read_list(data, 'NOT_enum', unparsed_keys, path)
    return _parse_enum(values, path, True)


def parse_enum(data: dict, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    values = _read_list(data, 'enum', unparsed_keys, path)
    return _parse_enum(values, path, False)


def _parse_enum(values: list, path: JsonPointer, invert: bool) -> Node:
    root = NoOpDecision(str(path), False)
    max_length = 0
    for value in values:
        root.add_transition(SetValueLeaf(None, not invert, value))
        max_length = max(max_length, len(str(value)))
    root.add_transition(SetValueLeaf(
        None, invert, "#" * (max_length+1)))
    return root


def parse_reference(data: dict, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    ref = _read_string(data, '$ref', unparsed_keys, path)
    return Reference(str(path), ref)


def parse_object(data: dict, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    props = _read_dict(data, 'properties', unparsed_keys, path, {})

    additional_props = _read_dict(data, 'additionalProperties', unparsed_keys, path, {})
    min_properties = _read_int(data, 'minProperties', unparsed_keys, path, 0)
    max_properties = _read_int(data, 'maxProperties', unparsed_keys, path, 0)
    pattern_properties = _read_dict(data, 'patternProperties', unparsed_keys, path, {})
    property_names = _read_dict(data, 'propertyNames', unparsed_keys, path, {})
    unevaluated_properties = _read_dict(data, 'unevaluatedProperties', unparsed_keys, path, {})
    depended_required = _read_dict(data, 'dependentRequired', unparsed_keys, path, [])
    dependent_schema = _read_dict(data, 'dependentSchemas', unparsed_keys, path, {})
    required = _read_list(data, 'required', unparsed_keys, path, [])

    required_props: Set[str] = set()
    for idx, token in enumerate(required):
        sub_path = path + 'required' + idx
        if not isinstance(token, str):
            raise JsonSchemaException(f"'{token}' is not a string at ${sub_path}", sub_path)
        if token in required_props:
            raise JsonSchemaException(f"Duplicate token '{token}' in ${sub_path}", sub_path)
        required_props.add(token)

    super_root = NoOpDecision(f"{path}_OBJECT")

    # Properties
    root = CreateObjectNode(None, True)
    super_root.add_transition(root)
    for key, value in props.items():
        sub_path = path + 'properties' + key

        property_root = NoOpDecision(f"{sub_path}__PROP", False)
        root.add_transition(property_root)

        # Insert property (always valid)
        key_node = InsertKeyNode(f"{sub_path}__KEY", key)
        property_root.add_transition(key_node)
        value_node = parse_dict(value, config, sub_path)
        key_node.add_transition(value_node)

        # Omit property (valid if not required)
        if key in required_props:
            property_root.add_transition(NoOpLeaf(None, is_valid=False))
            required_props.remove(key)
        else:
            property_root.add_transition(NoOpLeaf(None, is_valid=True))

    # All remaining properties only mentioned in 'required'
    for key in required_props:
        sub_path = path + 'required' + key
        property_root = NoOpDecision(f"{sub_path}__PROP", False)
        root.add_transition(property_root)
        key_node = InsertKeyNode(f"{sub_path}__KEY", key)
        property_root.add_transition(key_node)
        value_node = generate_default_samples(config)
        key_node.add_transition(value_node)

    if not root.outgoing_transitions:
        root.add_transition(NoOpLeaf(None, True))

    return super_root


def parse_string(data: dict, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    root = NoOpDecision()
    pattern = _read_string(data, 'pattern', unparsed_keys, path, '')
    content_media_type = _read_string(data, 'contentMediaType', unparsed_keys, path, '')
    content_encoding = _read_string(data, 'contentEncoding', unparsed_keys, path, '')
    content_schema = _read_dict(data, 'contentSchema', unparsed_keys, path, {})
    if pattern:
        regex = pattern
    else:
        regex = None
    # TODO: use format
    format = _read_string(data, 'format', unparsed_keys, path, None)
    if format is None:
        properties = StringProperties(
            min_length=_read_int(data, 'minLength', unparsed_keys, path, 0),
            max_length=_read_int(data, 'maxLength', unparsed_keys, path, float("inf")),
            # pattern=regex,
        )
        value = generate_random_string(properties)
    else:
        value = generate_random_format(format)
    root.add_transition(SetValueLeaf(None, True, value))
    return root

def generate_default_samples(config: Config) -> Node:
    items_node = NoOpDecision(None, False)
    for samples in config.default_samples.values():
        for sample in samples:
            items_node.add_transition(SetValueLeaf(None, True, sample))
    return items_node


def parse_array(data: dict, config: Config, unparsed_keys: Set[str], pointer: JsonPointer) -> Node:
    min_items = _read_int(data, 'minItems', unparsed_keys, pointer, 1)
    max_items = _read_int(data, 'maxItems', unparsed_keys, pointer, None)
    prefix_items = _read_list(data, 'prefixItems', unparsed_keys, pointer, {})
    unique_items = _read_dict(data, 'uniqueItems', unparsed_keys, pointer, {})
    contains = _read_dict(data, 'contains', unparsed_keys, pointer, None)
    min_contains = _read_int(data, 'minContains', unparsed_keys, pointer, 1)
    max_contains = _read_int(data, 'maxContains', unparsed_keys, pointer, None)

    items = _read_dict(data, 'items', unparsed_keys, pointer, None)

    root_node = CreateArrayNode(f"{pointer}_ARRAY", True)

    # Prefix items
    if prefix_items:
        prefix_items_node = NoOpDecision(f"{pointer}_PREFIX", True)
        root_node.add_transition(prefix_items_node)
        for idx, item in enumerate(prefix_items):
            node = parse_dict(item, config, pointer + 'prefixItems' + idx)
            append_node = AppendArrayItemNode(None)
            append_node.add_transition(node)
            prefix_items_node.add_transition(append_node)

    # Contained items
    if min_contains and contains is not None:
        contains_items_node = NoOpDecision(f"{pointer}_CONTAINS", True)
        root_node.add_transition(contains_items_node)
        node = parse_dict(contains, config, pointer + 'contains')
        append_node = AppendArrayItemNode(None)
        append_node.add_transition(node)
        for _ in range(min_contains):
            contains_items_node.add_transition(append_node)
        min_items = max(0, min_items - min_contains)

    # Items
    if min_items:
        all_items_node = NoOpDecision(f"{pointer}_ITEMS", True)
        root_node.add_transition(all_items_node)
        if items is None:
            items_node = generate_default_samples(config)
        else:
            items_node = parse_dict(items, config, pointer + 'items')

        for _ in range(min_items):
            append_node = AppendArrayItemNode(None)
            append_node.add_transition(items_node)
            all_items_node.add_transition(append_node)

    return root_node


def parse_boolean(data: dict, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    root = NoOpDecision(f"{path}_BOOLEAN")
    root.add_transition(SetValueLeaf(None, True, value=True))
    root.add_transition(SetValueLeaf(None, True, value=False))
    return root

def parse_number(data: dict, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    minimum = _read_float(data, 'minimum', unparsed_keys, path, None)
    minimum_exclusive = _read_float(data, 'exclusiveMinimum', unparsed_keys, path, None)
    maximum = _read_float(data, 'maximum', unparsed_keys, path, None)
    maximum_exclusive = _read_float(data, 'exclusiveMaximum', unparsed_keys, path, None)
    multiple_of = _read_float(data, 'multipleOf', unparsed_keys, path, None)

    if minimum_exclusive is not None:
        minimum = minimum_exclusive + 1  # TODO: actually just add one bit (https://docs.python.org/3/tutorial/floatingpoint.html)
    if maximum_exclusive is not None:
        maximum = maximum_exclusive - 1  # TODO: actually just subtract one bit

    invalid_values = []
    if minimum is not None:
        invalid_values.append(minimum - 1)
    if maximum is not None:
        invalid_values.append(maximum + 1)

    valid_value = minimum or maximum or 0
    if multiple_of is not None and abs(multiple_of) > 1e-5:
        valid_value = (int(valid_value / multiple_of)) * multiple_of
        if minimum is not None and valid_value < minimum:
            valid_value += multiple_of

    root = NoOpDecision(f"{path}_NUMBER")
    root.add_transition(SetValueLeaf(None, True, valid_value))
    for value in invalid_values:
        root.add_transition(SetValueLeaf(None, False, value))

    return root


def parse_null(data: dict, config: Config, unparsed_keys: Set[str], path: JsonPointer) -> Node:
    root = NoOpDecision(f"{path}_NULL")
    root.add_transition(SetValueLeaf(None, True, value=None))
    return root


def parse_any_of_entry(entry: dict, config: Config, pointer: JsonPointer) -> Node:
    unparsed_keys = set(entry.keys())

    # try special keys first
    for key, handler in config.key_handlers.items():
        if key in entry:
            return handler(entry, config, unparsed_keys, pointer)

    root = NoOpDecision(None, False)
    if 'type' in entry:
        types = entry['type']
        if isinstance(types, str):
            types = [types]
        types = set(types)
    else:
        # Omitting 'type' basically allows all types
        types = config.type_handlers.keys()

    # Generate samples for all allowed types (valid + invalid)
    for t in types:
        try:
            handler = config.type_handlers[t]
        except KeyError:
            raise JsonSchemaException(f"Unknown type '{t}'", pointer)
        root.add_transition(handler(entry, config, unparsed_keys, pointer))

    # Generate counter-examples for all forbidden types
    for type, samples in config.default_samples.items():
        if type not in types:
            for sample in samples:
                root.add_transition(SetValueLeaf(None, False, sample))

    # if unparsed_keys:
    #     raise JsonSchemaException(f"Unknown keys {unparsed_keys} at '{path}'", path)
    return root


def parse_dict(data: dict, config: Config, pointer: JsonPointer) -> Node:
    root = NoOpDecision(str(pointer), False)
    unparsed_keys = set(data.keys())
    any_of = _read_list(data, 'anyOf', unparsed_keys, pointer)

    for idx, entry in enumerate(any_of):
        root.add_transition(parse_any_of_entry(
            entry, config, pointer + 'anyOf' + idx
        ))

    return root


def parse(data: dict, config=None) -> Node:
    if config is None:
        config = default_config()

    if config.normalize:
        data = normalize(data)

    pointer = JsonPointer()
    unparsed_keys = set(data.keys())

    # Read definitions
    all_nodes: List[Node] = []
    definitions = _read_dict(data, '$defs', unparsed_keys, pointer, {})
    for key, definition in definitions.items():
        sub_path = pointer + '$defs' + key
        node = parse_dict(definition, config, sub_path)
        all_nodes.append(node)

    # Read root
    root = parse_dict(data, config, pointer)
    root = root.resolve(all_nodes)

    root.optimize()

    # prepend input / output nodes
    create_input = CreateInputNode()
    super_root = NoOpDecision(None, True)
    create_input.add_transition(super_root)
    super_root.add_transition(root)
    super_root.add_transition(FetchOutputNode())

    return create_input



from typing import Union
from fences.core.exception import JsonPointerException


class JsonPointer:

    def __init__(self, elements=None) -> None:
        self.elements = elements or []

    def __add__(self, other: Union[str, int]) -> "JsonPointer":
        if isinstance(other, str):
            return JsonPointer(self.elements + [other])
        if isinstance(other, int):
            return JsonPointer(self.elements + [str(other)])
        raise NotImplementedError()

    def __str__(self) -> str:
        return '#/' + '/'.join(self.elements)

    @classmethod
    def from_string(self, value: str) -> "JsonPointer":
        if value == '#/' or value == '#':
            return JsonPointer()

        if not value.startswith('#/'):
            raise JsonPointerException(f"Only local pointers are supported, got {value}")
        value = value[2:]
        return JsonPointer(value.split("/"))

    def is_root(self) -> bool:
        return len(self.elements) == 0

    def lookup(self, data: any, index: int = 0) -> any:
        if index > len(self.elements):
            raise JsonPointerException("Out of range")
        if index == len(self.elements):
            return data
        if isinstance(data, dict):
            key = self.elements[index]
            try:
                value = data[key]
            except KeyError:
                raise JsonPointerException(f"{key} not in data")
            return self.lookup(value, index+1)
        elif isinstance(data, list):
            try:
                i = int(self.elements[index])
            except ValueError:
                raise JsonPointerException(
                    f"{self.elements[index]} is not an integer for array lookup")
            try:
                value = data[i]
            except IndexError:
                raise JsonPointerException(f"Index not in array")
            return self.lookup(data[i], index+1)
        else:
            raise JsonPointerException(f"Cannot lookup in {data}")



from dataclasses import dataclass, field
from typing import Callable, Set, List, Dict

from .json_pointer import JsonPointer
from ..core.random import StringProperties

from fences.core.node import Decision

Handler = Callable[[dict, "Config", Set[str], JsonPointer], Decision]


@dataclass
class Config:
    key_handlers: Dict[str, Handler]
    type_handlers: Dict[str, Handler]
    default_samples: Dict[str, List[any]]
    normalize: bool



from fences.core.exception import ParseException
from .json_pointer import JsonPointer

class JsonSchemaException(ParseException):
    def __init__(self, message: str, json_pointer: JsonPointer) -> None:
        super().__init__(message)
        self.json_pointer = json_pointer



from .exception import ResolveReferenceException, InternalException
from typing import List, Optional, Generator, Set, Dict, Tuple
from dataclasses import dataclass

Path = List[int]


@dataclass
class ResultEntry:
    target: "Leaf"
    path: Path
    is_valid: bool


class Node:

    def __init__(self, id: Optional[str] = None) -> None:
        self.id = id
        self.incoming_transitions: List["IncomingTransition"] = []
        self._has_valid_successors: False
        self._has_invalid_successors: False

    def apply(self, data: any) -> any:
        """
        Applies to node's operation to the given data
        (Meant to be overridden)
        """
        raise NotImplemented()  # pragma: no cover

    def description(self) -> Optional[str]:
        """
        Returns a human readable description of the node' apply() function
        (Meant to be overridden)
        """
        return None

    def items(self) -> Generator["Node", None, None]:
        """
        Iterates over the node and all it's children recursively
        """
        already_visited: Set[Node] = set()
        yield from self._items(already_visited)

    def _items(self, already_visited: Set["Node"]) -> Generator["Node", None, None]:
        if id(self) in already_visited:
            return
        already_visited.add(id(self))
        yield self

    def target(self, nodes_by_id: Dict[str, "Node"]) -> "Node":
        """
        Returns the node's target, which is the referred node in case of a Reference or the node itself, otherwise
        """
        return self

    def resolve(self, nodes: List["Node"]) -> "Node":
        """
        Replace all Reference instances with the actual nodes they are pointing to.
        Returns the root - this is a new value if the root itself was a ReferenceNode.
        """

        # Collect references
        nodes_by_id: Dict[str, Node] = {}

        def insert(node: Node):
            if node.id and node.id in nodes_by_id:
                raise ResolveReferenceException(
                    f"id '{node.id}' exists multiple times", node.id)
            nodes_by_id[node.id] = node
        for node in nodes:
            for n in node.items():
                insert(n)
        for n in self.items():
            insert(n)

        # Get actual root
        root = self.target(nodes_by_id)

        # Resolve references
        visited = set()

        def _resolve(node: Node):
            if id(node) in visited:
                return
            visited.add(id(node))
            if isinstance(node, Decision):
                for idx, transition in enumerate(node.outgoing_transitions):
                    next_node = transition.target
                    if isinstance(next_node, Reference):
                        target_node = next_node.target(nodes_by_id)
                        transition.target = target_node
                        target_node.incoming_transitions.append(
                            IncomingTransition(node, idx)
                        )
                for transition in node.outgoing_transitions:
                    _resolve(transition.target)
        _resolve(root)
        return root

    def execute(self, path: Path, data=None) -> any:
        """
        Executes a path on the node
        """
        idx, result = self._execute(path, 0, data)
        if idx != len(path):  # pragma: no cover
            raise InternalException("Path not fully consumed")
        return result

    def _execute(self, path: Path, path_idx: int, data: any) -> Tuple[int, any]:
        if path_idx > len(path):  # pragma: no cover
            raise InternalException(
                f"path_idx ({path_idx}) > len(path) ({len(path)})")

        data = self.apply(data)

        if not isinstance(self, Decision):
            return path_idx, data

        if self.all_transitions:
            result = None
            for transition in self.outgoing_transitions:
                path_idx, result = transition.target._execute(
                    path, path_idx, data)
            # return the result of the last node
            return path_idx, result
        else:
            idx = path[path_idx]
            return self.outgoing_transitions[idx].target._execute(path, path_idx+1, data)

    def _generate(self, result_path: Path, already_reached: Set, on_valid_path: bool):
        already_reached.add(id(self))

        if isinstance(self, Leaf) and self.is_valid != on_valid_path:
            v = 'valid' if on_valid_path else 'invalid'
            raise InternalException(
                f"Decision without {v} leaf detected! {self.description()}")

        if not isinstance(self, Decision):
            return
        if not self.outgoing_transitions:
            return

        if self.all_transitions:
            for transition in self.outgoing_transitions:
                v = not on_valid_path if transition.is_inverting else on_valid_path
                transition._num_paths += 1
                transition.target._generate(result_path, already_reached, v)
        else:
            selected = None
            min_paths = float('inf')
            for idx, transition in enumerate(self.outgoing_transitions):
                target = transition.target
                if transition._num_paths < min_paths:
                    if on_valid_path and not target._has_valid_successors:
                        continue
                    if not on_valid_path and not target._has_invalid_successors:
                        continue
                    selected = idx
                    min_paths = transition._num_paths
            if selected is None:
                v = 'valid' if on_valid_path else 'invalid'
                raise InternalException(
                    f"Decision without {v} leaf detected ' (id={self.id})")

            result_path.append(selected)
            transition = self.outgoing_transitions[selected]
            transition._num_paths += 1
            v = not on_valid_path if transition.is_inverting else on_valid_path
            transition.target._generate(result_path, already_reached, v)

    def _backward(self, path: Path, already_reached: Set) -> Tuple[bool, "Node"]:
        already_reached.add(id(self))

        if not self.incoming_transitions:
            return False, self

        predecessor_transition = None
        min_paths = float('inf')
        for transition in self.incoming_transitions:
            n = transition.outgoing_transition()._num_paths
            if n < min_paths:
                min_paths = n
                predecessor_transition = transition
        path.append(predecessor_transition.outgoing_idx)
        predecessor_transition.outgoing_transition()._num_paths += 1
        is_inverting, root = predecessor_transition.source._backward(
            path, already_reached)
        if predecessor_transition.outgoing_transition().is_inverting:
            return not is_inverting, root
        else:
            return is_inverting, root

    def _forward(self, backward_path: Path, forward_path: Path, visited: Set, on_valid_path: bool):
        if len(backward_path) == 0:
            return
        assert isinstance(self, Decision)
        path_idx = backward_path.pop(-1)
        if self.all_transitions:
            for idx, transition in enumerate(self.outgoing_transitions):
                v = not on_valid_path if transition.is_inverting else on_valid_path
                if idx == path_idx:
                    transition.target._forward(
                        backward_path, forward_path, visited, v)
                else:
                    transition._num_paths += 1
                    transition.target._generate(forward_path, visited, v)
        else:
            transition = self.outgoing_transitions[path_idx]
            v = not on_valid_path if transition.is_inverting else on_valid_path
            forward_path.append(path_idx)
            transition._num_paths += 1
            transition.target._forward(backward_path, forward_path, visited, v)

    def _reset_stats(self, visited: Set[str]):
        if id(self) in visited:
            return
        visited.add(id(self))

        if isinstance(self, Leaf):
            self._has_valid_successors = self.is_valid
            self._has_invalid_successors = not self.is_valid
        else:
            assert isinstance(self, Decision)

            self._has_valid_successors = False
            self._has_invalid_successors = False
            for i in self.outgoing_transitions:
                i.target._reset_stats(visited)
                self._has_valid_successors = self._has_valid_successors or i.target._has_valid_successors
                self._has_invalid_successors = self._has_invalid_successors or i.target._has_invalid_successors

    def generate_paths(self) -> Generator[ResultEntry, None, None]:
        """
        Generates as many paths until all nodes in the graph are reached.
        Execute a path using execute().
        """

        valid_nodes: List[Leaf] = []
        invalid_nodes: List[Leaf] = []

        # Reset counter, collect leafs
        for node in self.items():
            if isinstance(node, Leaf):
                if node.is_valid:
                    valid_nodes.append(node)
                else:
                    invalid_nodes.append(node)
            if isinstance(node, Decision):
                for i in node.outgoing_transitions:
                    i._num_paths = 0

        to_visit = valid_nodes + invalid_nodes

        visited = set()
        self._reset_stats(visited)

        while to_visit:
            # print(f"{len(to_visit)} nodes remaining")
            next = to_visit[0]

            # Generate a path to the root
            backward_path: Path = []
            visited = set()
            is_inverting, root = next._backward(backward_path, visited)

            # Follow path to the target node
            forward_path = []
            root._forward(backward_path, forward_path, visited, True)

            # Yield
            is_valid = not next.is_valid if is_inverting else next.is_valid
            yield ResultEntry(next, forward_path, is_valid)

            # Remove the visited nodes
            path_idx = 0
            while path_idx < len(to_visit):
                if id(to_visit[path_idx]) in visited:
                    del to_visit[path_idx]
                else:
                    path_idx += 1

    def optimize(self):
        """
        Reduces the number of nodes in this graph while keeping the meaning the same.
        This helps to speed up subsequent operations.
        """
        pass

    def get_by_id(self, id: str) -> "Node":
        """
        Gets a specific node in the graph by it's id.
        Raises a KeyError if the id does not exist.
        """
        for i in self.items():
            if i.id == id:
                return i
        raise KeyError("No such id '{id}'")


class Leaf(Node):
    def __init__(self, id: str = None, is_valid: bool = True) -> None:
        super().__init__(id)
        self.is_valid = is_valid


class OutgoingTransition:
    def __init__(self, target: Node, is_inverting: bool) -> None:
        self.target = target
        self.is_inverting = is_inverting
        self._num_paths: int = 0


class IncomingTransition:
    def __init__(self, source: "Decision", idx: int) -> None:
        self.source = source
        self.outgoing_idx = idx

    def outgoing_transition(self) -> OutgoingTransition:
        return self.source.outgoing_transitions[self.outgoing_idx]


class Decision(Node):
    def __init__(self, id: str = None, all_transitions: bool = False) -> None:
        super().__init__(id)
        self.all_transitions = all_transitions
        self.outgoing_transitions: List[OutgoingTransition] = []

    def add_transition(self, target: Node, is_inverting: bool = False):
        target.incoming_transitions.append(
            IncomingTransition(self, len(self.outgoing_transitions))
        )
        self.outgoing_transitions.append(
            OutgoingTransition(target, is_inverting)
        )

    def _items(self, already_visited: Set["Node"]) -> Generator["Node", None, None]:
        if id(self) in already_visited:
            return
        already_visited.add(id(self))
        yield self
        for transition in self.outgoing_transitions:
            yield from transition.target._items(already_visited)

    def optimize(self):
        visited = set()
        self._optimize(visited)

    def _optimize(self, visited: Set):
        if id(self) in visited:
            return
        visited.add(id(self))

        merge_with = self
        is_inverting = False
        while True:
            if len(merge_with.outgoing_transitions) != 1:
                break
            successor = merge_with.outgoing_transitions[0].target
            if len(successor.incoming_transitions) != 1:
                break
            if not isinstance(successor, NoOpDecision):
                break
            if id(successor) in visited:
                break
            if merge_with.outgoing_transitions[0].is_inverting:
                is_inverting = not is_inverting
            merge_with = successor

        if merge_with is not self:
            self.outgoing_transitions = merge_with.outgoing_transitions
            self.all_transitions = merge_with.all_transitions
            if is_inverting:
                for i in self.outgoing_transitions:
                    i.is_inverting = not i.is_inverting
            for i in self.outgoing_transitions:
                for j in i.target.incoming_transitions:
                    if j.source is merge_with:
                        j.source = self

        for i in self.outgoing_transitions:
            if isinstance(i.target, Decision):
                i.target._optimize(visited)


class Reference(Node):
    def __init__(self, id: str, reference: str) -> None:
        super().__init__(id)
        self.reference = reference

    def target(self, nodes_by_id: Dict[str, "Node"]) -> Node:
        try:
            return nodes_by_id[self.reference].target(nodes_by_id)
        except KeyError as e:
            raise ResolveReferenceException(
                f"Unknown reference '{self.reference}'", self.reference)


class NoOpDecision(Decision):
    def apply(self, data: any) -> any:
        return data

    def description(self) -> Optional[str]:
        return "(Do nothing)"


class NoOpLeaf(Leaf):
    def apply(self, data: any) -> any:
        return data

    def description(self) -> Optional[str]:
        return "(do nothing)"
